@@ -4,9 +4,15 @@ import inspect
 import re
 import sys
 import textwrap
+import types
 from ast import FunctionDef, Module, stmt
+from dataclasses import dataclass
 from typing import Any, AnyStr, Callable, ForwardRef, NewType, TypeVar, get_type_hints
 
+from docutils.frontend import OptionParser
+from docutils.nodes import Node
+from docutils.parsers.rst import Parser as RstParser
+from docutils.utils import new_document
 from sphinx.application import Sphinx
 from sphinx.config import Config
 from sphinx.environment import BuildEnvironment
@@ -16,17 +22,38 @@ from sphinx.util import logging
 from sphinx.util.inspect import signature as sphinx_signature
 from sphinx.util.inspect import stringify_signature
 
+from .patches import install_patches
 from .version import __version__
 
 _LOGGER = logging.getLogger(__name__)
 _PYDATA_ANNOTATIONS = {"Any", "AnyStr", "Callable", "ClassVar", "Literal", "NoReturn", "Optional", "Tuple", "Union"}
 
+# types has a bunch of things like ModuleType where ModuleType.__module__ is
+# "builtins" and ModuleType.__name__ is "module", so we have to check for this.
+_TYPES_DICT = {getattr(types, name): name for name in types.__all__}
+# Prefer FunctionType to LambdaType (they are synonymous)
+_TYPES_DICT[types.FunctionType] = "FunctionType"
+
+
+def _get_types_type(obj: Any) -> str | None:
+    try:
+        return _TYPES_DICT.get(obj)
+    except Exception:
+        # e.g. exception: unhashable type
+        return None
+
 
 def get_annotation_module(annotation: Any) -> str:
     if annotation is None:
         return "builtins"
+    if _get_types_type(annotation) is not None:
+        return "types"
     is_new_type = sys.version_info >= (3, 10) and isinstance(annotation, NewType)
-    if is_new_type or isinstance(annotation, TypeVar) or type(annotation).__name__ == "ParamSpec":
+    if (
+        is_new_type
+        or isinstance(annotation, TypeVar)
+        or type(annotation).__name__ in ("ParamSpec", "ParamSpecArgs", "ParamSpecKwargs")
+    ):
         return "typing"
     if hasattr(annotation, "__module__"):
         return annotation.__module__  # type: ignore # deduced Any
@@ -35,17 +62,23 @@ def get_annotation_module(annotation: Any) -> str:
     raise ValueError(f"Cannot determine the module of {annotation}")
 
 
+def _is_newtype(annotation: Any) -> bool:
+    if sys.version_info < (3, 10):
+        return inspect.isfunction(annotation) and hasattr(annotation, "__supertype__")
+    else:
+        return isinstance(annotation, NewType)
+
+
 def get_annotation_class_name(annotation: Any, module: str) -> str:
     # Special cases
     if annotation is None:
         return "None"
-    elif annotation is Any:
-        return "Any"
-    elif annotation is AnyStr:
+    if annotation is AnyStr:
         return "AnyStr"
-    elif (sys.version_info < (3, 10) and inspect.isfunction(annotation) and hasattr(annotation, "__supertype__")) or (
-        sys.version_info >= (3, 10) and isinstance(annotation, NewType)
-    ):
+    val = _get_types_type(annotation)
+    if val is not None:
+        return val
+    if _is_newtype(annotation):
         return "NewType"
 
     if getattr(annotation, "__qualname__", None):
@@ -139,6 +172,9 @@ def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901 # t
     if module == "typing_extensions":
         module = "typing"
 
+    if module == "_io":
+        module = "io"
+
     full_name = f"{module}.{class_name}" if module != "builtins" else class_name
     fully_qualified: bool = getattr(config, "typehints_fully_qualified", False)
     prefix = "" if fully_qualified or full_name == class_name else "~"
@@ -168,18 +204,20 @@ def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901 # t
     elif full_name in ("typing.Union", "types.UnionType") and type(None) in args:
         if len(args) == 2:
             full_name = "typing.Optional"
+            role = "data"
             args = tuple(x for x in args if x is not type(None))  # noqa: E721
         else:
             simplify_optional_unions: bool = getattr(config, "simplify_optional_unions", True)
             if not simplify_optional_unions:
                 full_name = "typing.Optional"
+                role = "data"
                 args_format = f"\\[:py:data:`{prefix}typing.Union`\\[{{}}]]"
                 args = tuple(x for x in args if x is not type(None))  # noqa: E721
-    elif full_name == "typing.Callable" and args and args[0] is not ...:
+    elif full_name in ("typing.Callable", "collections.abc.Callable") and args and args[0] is not ...:
         fmt = [format_annotation(arg, config) for arg in args]
         formatted_args = f"\\[\\[{', '.join(fmt[:-1])}], {fmt[-1]}]"
     elif full_name == "typing.Literal":
-        formatted_args = f"\\[{', '.join(repr(arg) for arg in args)}]"
+        formatted_args = "\\[{}]".format(", ".join(f"``{arg!r}``" for arg in args))
     elif full_name == "types.UnionType":
         return " | ".join([format_annotation(arg, config) for arg in args])
 
@@ -254,7 +292,11 @@ def process_signature(
 
     obj = inspect.unwrap(obj)
     sph_signature = sphinx_signature(obj)
-    parameters = [param.replace(annotation=inspect.Parameter.empty) for param in sph_signature.parameters.values()]
+
+    if app.config.typehints_use_signature:
+        parameters = list(sph_signature.parameters.values())
+    else:
+        parameters = [param.replace(annotation=inspect.Parameter.empty) for param in sph_signature.parameters.values()]
 
     # if we have parameters we may need to delete first argument that's not documented, e.g. self
     start = 0
@@ -277,7 +319,10 @@ def process_signature(
             if not isinstance(method_object, (classmethod, staticmethod)):
                 start = 1
 
-    sph_signature = sph_signature.replace(parameters=parameters[start:], return_annotation=inspect.Signature.empty)
+    sph_signature = sph_signature.replace(parameters=parameters[start:])
+    if not app.config.typehints_use_signature_return:
+        sph_signature = sph_signature.replace(return_annotation=inspect.Signature.empty)
+
     return stringify_signature(sph_signature).replace("\\", "\\\\"), None
 
 
@@ -314,12 +359,18 @@ def get_all_type_hints(autodoc_mock_imports: list[str], obj: Any, name: str) ->
 
 _TYPE_GUARD_IMPORT_RE = re.compile(r"\nif (typing.)?TYPE_CHECKING:[^\n]*([\s\S]*?)(?=\n\S)")
 _TYPE_GUARD_IMPORTS_RESOLVED = set()
+_TYPE_GUARD_IMPORTS_RESOLVED_GLOBALS_ID = set()
 
 
 def _resolve_type_guarded_imports(autodoc_mock_imports: list[str], obj: Any) -> None:
-    if hasattr(obj, "__module__") and obj.__module__ not in _TYPE_GUARD_IMPORTS_RESOLVED:
+    if (hasattr(obj, "__module__") and obj.__module__ not in _TYPE_GUARD_IMPORTS_RESOLVED) or (
+        hasattr(obj, "__globals__") and id(obj.__globals__) not in _TYPE_GUARD_IMPORTS_RESOLVED_GLOBALS_ID
+    ):
         _TYPE_GUARD_IMPORTS_RESOLVED.add(obj.__module__)
         if obj.__module__ not in sys.builtin_module_names:
+            if hasattr(obj, "__globals__"):
+                _TYPE_GUARD_IMPORTS_RESOLVED_GLOBALS_ID.add(id(obj.__globals__))
+
             module = inspect.getmodule(obj)
             if module:
                 try:
@@ -327,7 +378,7 @@ def _resolve_type_guarded_imports(autodoc_mock_imports: list[str], obj: Any) ->
                 except (TypeError, OSError):
                     ...  # no source code => no type guards
                 else:
-                    for (_, part) in _TYPE_GUARD_IMPORT_RE.findall(module_code):
+                    for _, part in _TYPE_GUARD_IMPORT_RE.findall(module_code):
                         guarded_code = textwrap.dedent(part)
                         try:
                             with mock(autodoc_mock_imports):
@@ -468,14 +519,21 @@ def split_type_comment_args(comment: str) -> list[str | None]:
     return result
 
 
-def format_default(app: Sphinx, default: Any) -> str | None:
+def format_default(app: Sphinx, default: Any, is_annotated: bool) -> str | None:
     if default is inspect.Parameter.empty:
         return None
     formatted = repr(default).replace("\\", "\\\\")
-    if app.config.typehints_defaults.startswith("braces"):
-        return f" (default: ``{formatted}``)"
+
+    if is_annotated:
+        if app.config.typehints_defaults.startswith("braces"):
+            return f" (default: ``{formatted}``)"
+        else:  # other option is comma
+            return f", default: ``{formatted}``"
     else:
-        return f", default: ``{formatted}``"
+        if app.config.typehints_defaults == "braces-after":
+            return f" (default: ``{formatted}``)"
+        else:
+            return f"default: ``{formatted}``"
 
 
 def process_docstring(
@@ -554,18 +612,26 @@ def _inject_types_to_docstring(
     name: str,
     lines: list[str],
 ) -> None:
-    for arg_name, annotation in type_hints.items():
-        if arg_name == "return":
-            continue  # this is handled separately later
-        if signature is None or arg_name not in signature.parameters:
-            default = inspect.Parameter.empty
-        else:
-            default = signature.parameters[arg_name].default
+    if signature is not None:
+        _inject_signature(type_hints, signature, app, lines)
+    if "return" in type_hints:
+        _inject_rtype(type_hints, original_obj, app, what, name, lines)
+
+
+def _inject_signature(
+    type_hints: dict[str, Any],
+    signature: inspect.Signature,
+    app: Sphinx,
+    lines: list[str],
+) -> None:
+    for arg_name in signature.parameters:
+        annotation = type_hints.get(arg_name, None)
+
+        default = signature.parameters[arg_name].default
+
         if arg_name.endswith("_"):
             arg_name = f"{arg_name[:-1]}\\_"
 
-        formatted_annotation = format_annotation(annotation, app.config)
-
         insert_index = None
         for at, line in enumerate(lines):
             if _line_is_param_line_for_arg(line, arg_name):
@@ -575,45 +641,155 @@ def _inject_types_to_docstring(
                 insert_index = at
                 break
 
-        if insert_index is None and app.config.always_document_param_types:
+        if annotation is not None and insert_index is None and app.config.always_document_param_types:
             lines.append(f":param {arg_name}:")
             insert_index = len(lines)
 
         if insert_index is not None:
-            type_annotation = f":type {arg_name}: {formatted_annotation}"
+            if annotation is None:
+                type_annotation = f":type {arg_name}: "
+            else:
+                formatted_annotation = format_annotation(annotation, app.config)
+                type_annotation = f":type {arg_name}: {formatted_annotation}"
+
             if app.config.typehints_defaults:
-                formatted_default = format_default(app, default)
+                formatted_default = format_default(app, default, annotation is not None)
                 if formatted_default:
                     if app.config.typehints_defaults.endswith("after"):
                         lines[insert_index] += formatted_default
                     else:  # add to last param doc line
                         type_annotation += formatted_default
+
             lines.insert(insert_index, type_annotation)
 
-    if "return" in type_hints and not inspect.isclass(original_obj):
-        if what == "method" and name.endswith(".__init__"):  # avoid adding a return type for data class __init__
-            return
-        formatted_annotation = format_annotation(type_hints["return"], app.config)
-        insert_index = len(lines)
-        for at, line in enumerate(lines):
-            if line.startswith(":rtype:"):
-                insert_index = None
-                break
-            elif line.startswith(":return:") or line.startswith(":returns:"):
-                if " -- " in line and not app.config.typehints_use_rtype:
-                    insert_index = None
-                    break
-                insert_index = at
 
-        if insert_index is not None and app.config.typehints_document_rtype:
-            if insert_index == len(lines):  # ensure that :rtype: doesn't get joined with a paragraph of text
-                lines.append("")
-                insert_index += 1
-            if app.config.typehints_use_rtype or insert_index == len(lines):
-                lines.insert(insert_index, f":rtype: {formatted_annotation}")
-            else:
-                line = lines[insert_index]
-                lines[insert_index] = f":return: {formatted_annotation} --{line[line.find(' '):]}"
+@dataclass
+class InsertIndexInfo:
+    insert_index: int
+    found_param: bool = False
+    found_return: bool = False
+    found_directive: bool = False
+
+
+# Sphinx allows so many synonyms...
+# See sphinx.domains.python.PyObject
+PARAM_SYNONYMS = ("param ", "parameter ", "arg ", "argument ", "keyword ", "kwarg ", "kwparam ")
+
+
+def node_line_no(node: Node) -> int | None:
+    """
+    Get the 1-indexed line on which the node starts if possible. If not, return
+    None.
+
+    Descend through the first children until we locate one with a line number or
+    return None if None of them have one.
+
+    I'm not aware of any rst on which this returns None, to find out would
+    require a more detailed analysis of the docutils rst parser source code. An
+    example where the node doesn't have a line number but the first child does
+    is all `definition_list` nodes. It seems like bullet_list and option_list
+    get line numbers, but enum_list also doesn't. *shrug*.
+    """
+    while node.line is None and node.children:
+        node = node.children[0]
+    return node.line
+
+
+def tag_name(node: Node) -> str:
+    return node.tagname  # type:ignore[attr-defined,no-any-return]
+
+
+def get_insert_index(app: Sphinx, lines: list[str]) -> InsertIndexInfo | None:
+    # 1. If there is an existing :rtype: anywhere, don't insert anything.
+    if any(line.startswith(":rtype:") for line in lines):
+        return None
+
+    # 2. If there is a :returns: anywhere, either modify that line or insert
+    #    just before it.
+    for at, line in enumerate(lines):
+        if line.startswith((":return:", ":returns:")):
+            return InsertIndexInfo(insert_index=at, found_return=True)
+
+    # 3. Insert after the parameters.
+    # To find the parameters, parse as a docutils tree.
+    settings = OptionParser(components=(RstParser,)).get_default_values()
+    settings.env = app.env
+    doc = new_document("", settings=settings)
+    RstParser().parse("\n".join(lines), doc)
+
+    # Find a top level child which is a field_list that contains a field whose
+    # name starts with one of the PARAM_SYNONYMS. This is the parameter list. We
+    # hope there is at most of these.
+    for child in doc.children:
+        if tag_name(child) != "field_list":
+            continue
+
+        if not any(c.children[0].astext().startswith(PARAM_SYNONYMS) for c in child.children):
+            continue
+
+        # Found it! Try to insert before the next sibling. If there is no next
+        # sibling, insert at end.
+        # If there is a next sibling but we can't locate a line number, insert
+        # at end. (I don't know of any input where this happens.)
+        next_sibling = child.next_node(descend=False, siblings=True)
+        line_no = node_line_no(next_sibling) if next_sibling else None
+        at = line_no - 2 if line_no else len(lines)
+        return InsertIndexInfo(insert_index=at, found_param=True)
+
+    # 4. Insert before examples
+    # TODO: Maybe adjust which tags to insert ahead of
+    for child in doc.children:
+        if tag_name(child) in ["literal_block", "paragraph", "field_list"]:
+            continue
+        line_no = node_line_no(child)
+        at = line_no - 2 if line_no else len(lines)
+        return InsertIndexInfo(insert_index=at, found_directive=True)
+
+    # 5. Otherwise, insert at end
+    return InsertIndexInfo(insert_index=len(lines))
+
+
+def _inject_rtype(
+    type_hints: dict[str, Any],
+    original_obj: Any,
+    app: Sphinx,
+    what: str,
+    name: str,
+    lines: list[str],
+) -> None:
+    if inspect.isclass(original_obj) or inspect.isdatadescriptor(original_obj):
+        return
+    if what == "method" and name.endswith(".__init__"):  # avoid adding a return type for data class __init__
+        return
+    if not app.config.typehints_document_rtype:
+        return
+
+    r = get_insert_index(app, lines)
+    if r is None:
+        return
+
+    insert_index = r.insert_index
+
+    if not app.config.typehints_use_rtype and r.found_return and " -- " in lines[insert_index]:
+        return
+
+    formatted_annotation = format_annotation(type_hints["return"], app.config)
+
+    if r.found_param and insert_index < len(lines) and lines[insert_index].strip() != "":
+        insert_index -= 1
+
+    if insert_index == len(lines) and not r.found_param:
+        # ensure that :rtype: doesn't get joined with a paragraph of text
+        lines.append("")
+        insert_index += 1
+    if app.config.typehints_use_rtype or not r.found_return:
+        line = f":rtype: {formatted_annotation}"
+        lines.insert(insert_index, line)
+        if r.found_directive:
+            lines.insert(insert_index + 1, "")
+    else:
+        line = lines[insert_index]
+        lines[insert_index] = f":return: {formatted_annotation} --{line[line.find(' '):]}"
 
 
 def validate_config(app: Sphinx, env: BuildEnvironment, docnames: list[str]) -> None:  # noqa: U100
@@ -634,10 +810,13 @@ def setup(app: Sphinx) -> dict[str, bool]:
     app.add_config_value("typehints_defaults", None, "env")
     app.add_config_value("simplify_optional_unions", True, "env")
     app.add_config_value("typehints_formatter", None, "env")
+    app.add_config_value("typehints_use_signature", False, "env")
+    app.add_config_value("typehints_use_signature_return", False, "env")
     app.connect("env-before-read-docs", validate_config)  # config may be changed after “config-inited” event
     app.connect("autodoc-process-signature", process_signature)
     app.connect("autodoc-process-docstring", process_docstring)
-    return {"parallel_read_safe": True}
+    install_patches(app)
+    return {"parallel_read_safe": True, "parallel_write_safe": True}
 
 
 __all__ = [
